C Program DistPCoA --
C This program performs Principal Coordinate Analysis (PCoA; Gower 1966) with
C correction for negative eigenvalues. This procedure is used as part of the 
C distance-based redundancy analysis method (db-RDA) proposed by Legendre & 
C Anderson (1998).
C 
C The program can either read in a pre-computed distance matrix, or calculate
C a distance matrix from a (site x species) data table. Five distance functions
C are available within the program: Bray-Curtis, sqrt(Bray-Curtis), chi-square,
C Hellinger, and Euclidean. A variety of transformations of the raw data are
C also available from within the program.
C 
C The program uses a Householder procedure for finding the eigenvalues and 
C eigenvectors of a square distance matrix. The subroutines (TRED2, TQLI) are
C from Chapter 11 of Numerical Recipes (Press et al., 1986).
C 
C Negative eigenvalues may be generated by the semimetric Bray-Curtis distance
C function, for example. This function is widely used in ecology to measure the
C distance between sites using species abundance data. Corrections for negative
C eigenvalues can be obtained using three methods:
C 
C 1. Lingoes method: d'(i,j)**2 = d(i,j)**2 + 2*c1   where c1 is the absolute  
C    value of the largest negative eigenvalue of the first PCoA run. Note that
C    d(i,i) = 0.
C 2. Cailliez method: d'(i,j) = d(i,j) + c2   where c2 is the largest eigenvalue
C    of a special non-symmetric matrix. Note that d(i,i) = 0. The eigenvalues of
C    the special matrix are found using a QR algorithm for real Hessenberg 
C    matrices. The subroutines (BALANC, ELMHES, and HQR) are from Chapter 11 of
C    Numerical Recipes (Press et al., 1986).
C 3. d'(i,j) = sqrt(d(i,j)).
C
C Methods 1 and 2 are described in Gower & Legendre (1986, theorem 7) and also 
C in Legendre & Anderson (1998); this program corrects typos in Gower & Legendre
C (1986) following Cailliez (1983). Method 3 is from Legendre & Anderson (1998).
C
C In the main program, 
C    pmax is the maximum number of objects and variables
C    p = no. of replicates (i.e. objects)
C    d = original distance matrix (does not change with each run)
C    alpha = Gower-centred matrix, then eigenvectors
C            (This matrix, alpha, changes with each run of the program,
C            when negative eigenvalues occur)
C    e = diagonal elements in tridiagonal matrix, then eigenvalues
C    e2= real portion of eigenvalues of Special matrix
C    g2= imaginary portion of eigenvalues of Special matrix
C
C                                  Pierre Legendre and Marti Jane Anderson, 1998
C
C 345678901234567890123456789012345678901234567890123456789012345678901234567890
      INTEGER p,p2,pmax,ppmax,nspec,run,ndis
      PARAMETER (pmax=400,ppmax=pmax*2)
      DIMENSION C(pmax,pmax),d(pmax,pmax),alpha(pmax,pmax),
     +          special(ppmax,ppmax),delta1(pmax,pmax),delta2(pmax,pmax)
      REAL*8 C,d,e2(ppmax),g2(ppmax),dconst2,AA,ddmin,ddmin1,corr,
     +       e(pmax),g(pmax),toteigv,dconst,trace,alpha,epsilon,
     +       sumcol(pmax),special,delta1,delta2,totc(pmax),totr(pmax)
      INTEGER number(pmax),number2(ppmax),nn,unt
      EQUIVALENCE (C,alpha), (delta1(1,1),delta2(1,1),special(1,1))
      CHARACTER tt*10
      Character*79 namea
C
C BBEdit is the 'creator' for output files
C     call F_Creator('R*ch')
C
      write(*,*) 'Principal coordinate analysis'
      write(*,*) 'with correction for negative eigenvalues, if any.'
      write(*,103) pmax
      write(*,*)
      
C Mod. PhC 17/05/00 to write results to the screen or in a file
      write(*,*) 'Do you wish to print results'
      write(*,*) '   (0) On screen'
      write(*,*) '   (1) In a file'
      read(*,*) iyesno
      if(iyesno.gt.0) then
C Write to a file, name it unit 7
          unt=7
          write(*,*)
          write(*,*) 'Name of output file?'
          read(*,*) namea
          open (unt,file=namea,status='new')
          inquire(unit=unt,name=namea)
          write(*,*) 'Results will be written to file ',namea
          write(*,*)
       else
C Screen unit
          unt=6
        endif

2     write(*,*) 
     +'Do you have a file with (1) a square Distance or Similarity ',
     +'matrix, or'
      write(*,*) '                        (2) raw data    ?'
      write(*,*) '(Type -1 or -2 to get intermediate matrices printed.)'
      read(*,*) ndata
      if((ndata.lt.-2).or.(ndata.gt.2).or.(ndata.eq.0)) goto 2
      goto (4,10) iabs(ndata)

4     write(*,*) 'Name of input file with square D or S matrix?'
      read(*,*) namea
      open (2,file=namea)
      inquire(unit=2,name=namea)
      write(*,*) 'Input file name (D or S matrix): ',namea
      write(*,*)
      write(*,*) 'How many objects?'
      read (5,*) p
      if(p.gt.pmax) stop 'Too many objects. Recompile the program.'
      read (2,*) ((d(i,j), j=1,p), i=1,p)
      close (2)
      write(*,*) 'Is it a Distance (0) or a Similarity (1) matrix ?'
      read(*,*) irep
      if (irep.eq.1) then
         do 6 i = 1,p
         do 6 j = 1,p
6        d(i,j) = 1.0 - d(i,j)
         endif
      do 8 i=1,p
8     d(i,i)=0.0
      goto 16

10    write(*,*) 'Name of input file with raw data?'
      write(*,*) '(in which columns are variables and rows are ',
     +'replicates)'
      read(*,*) namea
      open (1,file=namea)
      inquire(unit=1,name=namea)
      write(*,*) 'Input file name (raw data): ',namea
      write(*,*)
      write(*,*) 'How many objects?'
      read(5,*) p
      if(p.gt.pmax) stop 'Too many objects: recompile program.'
      write(*,*)
      write(*,*) 'How many variables?'
      read (5,*) nspec
      if(nspec.gt.pmax) stop 'Too many variables: recompile program'
C
      write(*,*)
      write(*,*) 'Transform the raw data before computing distances?'
      write(*,*) '   (0) No transformation'
      write(*,*) '   (1) y'' = sqrt(y),        i.e. y'' = y^0.5'
      write(*,*) '   (2) y'' = double sqrt(y), i.e. y'' = y^0.25'
      write(*,*) '   (3) y'' = ln(y)'
      write(*,*) '   (4) y'' = ln(y + 1)'
      write(*,*) '   (5) y'' = log10(y)'
      write(*,*) '   (6) y'' = log10(y + 1)'
      read(*,*) itr
      read (1,*) ((C(i,j), j = 1,nspec), i = 1,p)
      close (1)
      if(ndata.lt.0) then
         write(unt,*)
         write(unt,*) 'Raw data:'
         do 12 i=1,p
12       write(unt,101) (C(i,j), j=1,nspec)
         endif
      if(itr.gt.0) call Transform(C,p,pmax,nspec,ndata,itr)
C
C Choose a distance function
14    write(*,*)
      write(*,*) 'Options: (1) Bray-Curtis distance'
      write(*,*) '         (2) sqrt(Bray-Curtis)'
      write(*,*) '         (3) Chi-square distance'
      write(*,*) '         (4) Hellinger distance'
      write(*,*) '         (5) Euclidean distance'
      read (5,*) ndis
      if((ndis.lt.1).or.(ndis.gt.5)) then
         write(*,*) 'Incorrect answer. Try again...'
         goto 14
         endif
C
C Choose method of correction for negative eigenvalues
16    write(*,*) 'Correction for negative eigenvalues, if any: '
      write(*,*) "1) Method 1 (Lingoes):  d'(i,j) = sqrt(d(i,j)**2 ",
     +" 2*c1)"
      write(*,*) "2) Method 2 (Cailliez): d'(i,j) = d(i,j) + c2"
      write(*,*) "3) No correction: yields coordinates corresponding"
      write(*,*) "   to positive eigenvalues only"
      read (5,*) icorr
      if((icorr.lt.1).or.(icorr.gt.3)) then
         write(*,*) 'Incorrect answer. Try again...'
         goto 16
         endif
C
C     call time(tt)
      write(*,*) tt
      debut=secnds(0.0)
      write(*,*)
      run = 0
      dconst  = 0.0
      dconst2 = 0.0
      p2=p+p
      if (iabs(ndata).eq.2) 
     +   call Distance(C,d,totc,totr,p,pmax,nspec,ndis)

C The matrix is now square, with 0's on the main diagonal. Print it.
      if(ndata.lt.0) then
         write(unt,*)
         write(unt,*) 'Original matrix D'
         do 18 i=1,p
18       write(unt,101) (d(i,j), j=1,p)
         endif
C -------
C The following section of code is deactivated. It allows users to modify a
C matrix of Euclidean distances to make it uniformly non-Euclidean. This is 
C a way to test that the program will find the right correction value to make
C the matrix Euclidean again. To reactivate this section, remove the statement
C "goto 26" and recompile the program.
      goto 26
      if(ndis.eq.5) then   
         write(*,*)      
         write(*,*) 'Change Euclidean distance matrix ',
     +              'to make it uniformly non-Euclidean?'
         write(*,107) ddmin
         if(icorr.eq.1) then
            ddmin1=0.5*ddmin**2
            else
            ddmin1=ddmin
            endif
         write(*,108) ddmin1
20       write(*,*) 'Change it by how much? (0 = don''t do it)'
         read(5,*) corr
         if((corr.lt.0.0).or.(corr.gt.ddmin)) then
            write(*,*) 'Incorrect answer. Try again...'
            goto 20
            endif
         if(ddmin.gt.0.0) then
            do 22 i=1,p-1
            do 22 j=i+1,p
            if(icorr.eq.1) then
               d(i,j)=dsqrt(d(i,j)**2-2.0*corr)
               else
               d(i,j)=d(i,j)-corr
               endif
            d(j,i)=d(i,j)
22          continue
            if((ndata.lt.0).and.(corr.gt.0.0)) then
               write(unt,*)
               write(unt,*) 'Modified, non-Euclidean matrix D'
               do 24 i=1,p
24             write(unt,101) (d(i,j), j=1,p)
               endif
            endif
         endif
C -------
C Preparing for Principal coordinate analysis
26    run = run + 1
      write (unt,*)
      if (run.eq.1) then
         write (unt,*)
     +   '*** Results of PCoA on the original distance matrix ***'
         ELSE
         write (unt,*)'*** Results of PcoA on corrected distance ',
     +   'matrix ***'
         endif
C
C Method 1 (Lingoes): d'(i,j)**2 = d(i,j)**2 + 2*c1   where c1 is the 
C ABSOLUTE VALUE of the largest negative eigenvalue of the first PCoA run.
C This is obtained by computing alpha(i,j) = -0.5*d(i,j)**2 + dconst
C where dconst is the SIGNED largest negative eigenvalue of the first PCoA run.
C
C Method 2 (Cailliez): d'(i,j) = d(i,j) + c2   where c2 is the largest
C eigenvalue (dconst2) of a special non-symmetric matrix.
C
      if(run.eq.1) then
         do 28 i=1,p
         do 28 j=1,p
28       alpha(i,j)= -0.5*(d(i,j)**2)
      else
         if (icorr.eq.1) then
            do 30 i=1,p
            do 30 j=1,p
30          alpha(i,j) = -0.5*d(i,j)**2 + dconst
         else
            do 32 i=1,p
            do 32 j=1,p
32          alpha(i,j)=-0.5*(d(i,j)+dconst2)**2
         end if 
      endif
C Put zeros on diagonal of matrix alpha
      do 34 i=1,p
34    alpha(i,i)=0.0
      if(ndata.lt.0) then
         write(unt,*)
         write(unt,*) 'Matrix A = [-0.5 d**2]'
         do 36 i=1,p
36       write(unt,101) (alpha(i,j), j=1,p)
         endif
C
C Center matrix alpha in the manner of Gower
      call Gower(alpha,sumcol,p,pmax)
      if(ndata.lt.0) then
         write(unt,*)
         write(unt,*) 'Matrix alpha (= matrix A, Gower-centred)'
         do 38 i=1,p
38       write(unt,101) (alpha(i,j), j=1,p)
         endif
      do 40 i=1,p
      do 40 j=1,p
40    delta1(i,j)= 2.0*alpha(i,j)
      trace=0.0
      do 42 i=1,p
42    trace=trace+alpha(i,i)
      write(unt,*)
      write(unt,110) trace
C
C Calculate eigenvalues of Gower-centred alpha for PCoA
C TQLI returns the eigenvectors in matrix alpha
      if(run.eq.2) then
         do 44 i=1,p
            e(i) = 0.0
44          g(i) = 0.0
      end if
      call tred2(alpha,p,pmax,e,g)
      call tqli(e,g,p,pmax,alpha)
C
C Put the eigenvalues and corresponding eigenvectors in descending order
C Phase 1: sort eigenvalues and vector of integers "number"
      do 46 i=1,p
46    number(i)=i
      do 52 j = 2,p
            AA = e(j)
            nn=number(j)
            do 48 i = j-1,1,-1
                  if (e(i).gt.AA) goto 50
                  e(i+1) = e(i)
                  number(i+1)=number(i)
48                continue
            i = 0
50          e(i+1) = AA
            number(i+1)=nn
52    continue
C
C Phase 2: Sort "alpha" according to new order of "number". Use "totc" as buffer
      do 58 ii=1,p
      do 54 j=1,p
54    totc(j)=alpha(ii,j)
      do 56 j=1,p
56    alpha(ii,j)=totc(number(j))
58    continue
C
C Normalize the principle coordinates to variance=eigenvalue
      do 62 k=1,p
      if(e(k).gt.0.0) then
         temp=dsqrt(e(k))
         else
         temp=dsqrt(-e(k))
         endif
      do 60 i=1,p
60    alpha(i,k)=alpha(i,k)*temp
62    continue
C
C Print results
      write(unt,*)
      write(unt,102)
      write(unt,101) (e(j), j=1,p)
C Check to see if there are any negative eigenvalues:
      num=0.0
      do 64 i = 1,p
      if (e(i).lt.-0.000005) num=num+1
64    continue
      if ((num.eq.0).and.(run.eq.1)) then
         write(unt,*)
         write(unt,*) 'There are no negative eigenvalues.'
         write(unt,*) 'No correction is necessary.'
         run = run+1
         goto 68
         else if ((run.eq.1).and.(icorr.eq.3)) then
         write(unt,*) 'The negative eigenvalues, if any,'
         write(unt,*) 'are being ignored in this analysis.'
         run = run+1
         goto 68
         endif
C By changing the order, the largest negative eigenvalue
C is now found at the end of the list, but it is not necessarily
C e(p), because there may be fewer eigenvalues than replicates!
      If ((run.eq.1).and.(icorr.eq.1)) then
         do 66 i = 1,p
         if (e(i).lt.dconst) dconst = e(i)
66       continue
         write(unt,104) dconst
         endif
68    continue

C Resume the program
      toteigv=0.0
      do 70 j=1,p
70    toteigv=toteigv+e(j)
      write(unt,*)
      write (unt,109) toteigv
      write(unt,*)
      test=toteigv-trace
      if(abs(test).gt.0.00001) then
         write(unt,*) 'Warning: The sum of eigenvalues does not ',
     +   'equal the'
         write(unt,*) '         trace of the Gower-centred matrix ',
     +   'precisely.'
         write(unt,*)
         end if

C Method 2: Create the special matrix and find its largest eigenvalue
      if((run.eq.1).and.(icorr.eq.2)) then 
      write(unt,*) '*** Create Special matrix and find its largest ',
     +'eigenvalue ***'
C        Form matrix Delta1 and copy it into Special
         if(ndata.lt.0) then
            write(unt,*)
            write(unt,*) 'Matrix delta1 (pxp) = 2 alpha (Gower-centred)'
            do 72 i=1,p
72          write(unt,101) (delta1(i,j), j=1,p)
            endif
         do 74 i=1,p
         do 74 j=p+1,p2
74       special(i,j)=delta1(i,j-p)
C        Create matrix  delta2 = (-1/2)*d
         do 76 i=1,p
         do 76 j=1,p
76       delta2(i,j)=-0.5*d(i,j)
         if(ndata.lt.0) then
            write(unt,*)
            write(unt,*) 'Matrix delta2 = [-0.5 d]'
            do 78 i=1,p
78          write(unt,101) (delta2(i,j), j=1,p)
            endif
         call Gower(delta2,sumcol,p,pmax)
         if(ndata.lt.0) then
            write(unt,*)
            write(unt,*) 'Matrix delta2 (Gower-centred)'
            do 80 i=1,p
80          write(unt,101) (delta2(i,j), j=1,p)
            endif
C        Compute  delta2 = -4.0*delta2  and copy it into Special
         do 82 i=1,p
         do 82 j=1,p
            delta2(i,j)= -4.0*delta2(i,j)
82       continue
         if(ndata.lt.0) then
            write(unt,*)
            write(unt,*) 'Matrix delta2 (pxp) = -4 delta2 ',
     +      '(Gower-centred)'
            do 84 i=1,p
84          write(unt,101) (delta2(i,j), j=1,p)
            endif
         do 86 i=p+1,p2
         do 86 j=p+1,p2
            special(i,j)=delta2(i-p,j-p)
86       continue
C        Write submatrices 0 and -I into Special
         do 88 i=1,p
         do 88 j=1,p
            special(i,j)=0.0
88       continue
         do 90 i=p+1,p2
         do 90 j=1,p
            if(i-p.ne.j) then
               special(i,j)=0.0
            else
               special(i,j)=-1.0
            end if
90       continue
      if(ndata.lt.0) then
         write(unt,*)
         write(unt,*) 'Special matrix (2px2p)'
         do 92 i=1,p2
92       write(unt,101) (special(i,j), j=1,p2)
         endif
C Calculate the eigenvalues for the Special matrix
C Procedures (Numerical recipes) to find eigenvalues of real nonsymmetric matrix
         call Balanc(special,p2,ppmax)
         call Elmhes(special,p2,ppmax)
         call Hqr(special,p2,ppmax,e2,g2)
C Hqr does not necessarily return the eigenvalues in decreasing order. We now go
C through the list of eigenvalues to identify the largest and put it in dconst2.
         ndata2=0
         do 94 i=1,p2
         if(e2(i).gt.dconst2) then
            dconst2=e2(i)
            ii=i
            endif
94       continue
         if(g2(ii).ne.0.0) then
           ndata2=1
           write(unt,*)
           write(unt,*)'WARNING: the largest eigenvalue of the ',
     +     'Special matrix'
           write(unt,*)'is complex. This may not be the correction ',
     +     'that you want.'
           write(unt,*) 
     +     'Here is the list of eigenvalues (real and imaginary ',
     +     'portions).'
           endif
         if((ndata.lt.0).or.(ndata2.eq.1)) then
            write(unt,*)
            write(unt,*) 'Real portion of eigenvalues of Special matrix'
            write(unt,101) (e2(i), i=1,p2)
            write(unt,*)
            write(unt,*) 'Imaginary portion of eigenvalues of ',
     +      'Special matrix'
            write(unt,101) (g2(i), i=1,p2)
            endif
         write(unt,105) dconst2
      end if
C End of calculations on Special matrix

      write(*,*)
C     call time(tt)
      write(*,*) tt
C Calculating time elapsed
      fin=secnds(debut)
      write(*,106) fin

C The following checks for zero eigenvalues according to the threshold
C set by the value of epsilon, and prints only the non-zero coordinates.
      if (run.eq.2) then
         epsilon=0.000005
         nthresh=0
         do 96 i=p,1,-1
            if(icorr.eq.3) then
		       if (e(i).le.epsilon) then
                  nthresh=nthresh+1
                  cycle
                  else
                  goto 97
                  end if
		    else
		       if (dabs(e(i)).le.epsilon) then
                  nthresh=nthresh+1
                  cycle
                  else
                  goto 97
                  end if
	        end if
96       continue
97       continue
C Modification by PhC 17/05/00
C Asks for the output file name for the pcoord matrix
         write(*,*) 
     +   'Do you wish to save the principal coordinates in a ',
     +   'separate file?'
         write(*,*) '   (0) No'
         write(*,*) '   (1) Yes'
         read(*,*) iyesno
         if(iyesno.gt.0) then
            if(icorr.ne.3) then
               write(unt,*) 'The number of non-zero eigenvalues is:',
     +         p-nthresh
               else
               write(unt,*) 'The number of positive eigenvalues is:',
     +         p-nthresh
               end if
            write(*,*)
            write(*,*) 'Name of coordinates file?'
            read(*,*) namea
            open (9,file=namea,status='new')
            inquire(unit=9,name=namea)
            if(icorr.ne.3) then
               write(unt,*) 'Non-zero Principal coordinates'
               else
               write(unt,*)'Principal coordinates corresponding'
               write(unt,*)'to positive eigenvalues only'
               end if
            do 98 i=1,p
            write (9,101) (alpha(i,jj), jj=1,p-nthresh)
98          continue
            close (9)
            write(unt,*) 'have been written to output file: "',namea,'"'
         end if
         goto 99
      else
         goto 26
      end if
C
99    write(*,*)
      write(*,*) 'End of program.'
C Mod. PhC 17/05/00
C Close the output file if it's not the screen
      if(unt.ne.6) then
        close (unt)
        endif
      stop
101   format(1x,8f10.5)
102   format(' PCoA eigenvalues')
103   format(' Maximum size of matrix:',i5,' objects and descriptors')
104   format(/' The largest negative eigenvalue is',f15.10)
105   format(/' The largest eigenvalue of the Special matrix is',f15.10)
106   format(' Real time spent:',f10.2,' seconds')
107   format(' The smallest distance in matrix D is ',f10.5)
108   format(' Use a correction value smaller than  ',f10.5)
109   format(' Sum of computed eigenvalues =           ',f10.5)
110   format(' Trace of Gower-centred matrix =         ',f10.5)
      end

      SUBROUTINE Gower(A,sumcol,p,pmax)
C This subroutine centers the data in the manner of Gower (1966).
      INTEGER p,pmax
      Real*8 A(pmax,pmax),sumcol(pmax),supmax,pp
      pp=dfloat(p)
      do 2 i=1,p
      sumcol(i)=0.0
2     continue
      do 4 i=1,p
      do 4 j=1,p
4     sumcol(j)=sumcol(j)+A(i,j)
      supmax=0.
      do 6 j=1,p
      supmax=supmax+sumcol(j)
6     sumcol(j)=sumcol(j)/pp
      supmax=supmax/pp
      supmax=supmax/pp
      do 8 i=1,p
      do 8 j=1,p
8     A(i,j)=A(i,j)-sumcol(i)-sumcol(j)+supmax
      return
      end

      SUBROUTINE Transform(C,p,pmax,nspec,ndata,itr)
      INTEGER p,pmax
      Real*8 C(pmax,pmax)
C This subroutine computes the data transformation selected in the main program
      do 14 i=1,p
      do 14 j=1,nspec
      goto (2,4,6,8,10,12) itr
2     if(C(i,j).lt.0.0) goto 98
      C(i,j)=dsqrt(C(i,j))
      goto 14
4     if(C(i,j).lt.0.0) goto 98
      C(i,j)=C(i,j)**0.25
      goto 14
6     if(C(i,j).le.0.0) goto 99
      C(i,j)=dlog(C(i,j))
      goto 14
8     if((C(i,j)+1.0).le.0.0) goto 99
      C(i,j)=dlog(C(i,j)+1.0)
      goto 14
10    if(C(i,j).le.0.0) goto 99
      C(i,j)=dlog10(C(i,j))
      goto 14
12    if((C(i,j)+1.0).le.0.0) goto 99
      C(i,j)=dlog10(C(i,j)+1.0)
14    continue
      if(ndata.lt.0) then
         write(*,*)
         write(*,*) 'Transformed data:'
         do 16 i=1,p
16       write(*,101) (C(i,j), j=1,nspec)
         endif
      return
98    stop 'Cannot take the root of negative values.'
99    stop 'Cannot log-transform zero or negative values.'
101   format(1x,8f10.5)
      end

      SUBROUTINE Distance(C,d,totc,totr,p,pmax,nspec,ndis)
      INTEGER p,pmax
      Real*8 A,B,C(pmax,pmax),d(pmax,pmax),totc(pmax),totr(pmax),
     +       sumr,sumc,tott
C This subroutine computes the distance function selected in the main program
C Which method / distance measure?
      goto (34,34,40,40,62) ndis
C
C Calculation for Bray-Curtis and dsqrt(Bray-Curtis) distances
34    do 38 i = 1,p-1
         do 38 j = i+1,p
            A = 0.0
            B = 0.0
            do 36 k = 1, nspec
               A = A + dabs(C(i,k) - C(j,k))
               B = B + (C(i,k) + C(j,k))
36          continue
            If (B.lt.0.00000001) then
               d(i,j)=1.0
               else
               d(i,j) = A/B
               if(ndis.eq.2) d(i,j)=dsqrt(d(i,j))
               end if
38    continue
      goto 68
C
C Calculation for Chi-square and Hellinger distances
40    tott=0.0
      do 44 k = 1,p
         sumr = 0.0
         do 42 k2 = 1,nspec
            sumr = sumr + C(k,k2)
42       continue
         totr(k) = sumr
44       tott=tott+sumr
      do 48 k = 1,nspec
         sumc = 0.0
         do 46 k2 = 1,p
            sumc = sumc + C(k2,k)
46       continue
48       totc(k) = sumc
      goto (50,56) (ndis-2)
C Chi-square distance (NOT the chi-square metric)
C If a species is absent from the whole data file, skip it
50    do 54 i = 1,p-1
         do 54 j = i+1,p
            A = 0.0
            do 52 k = 1,nspec
               if (totc(k).eq.0.0) goto 52
               A = A+(1.0/totc(k))*(C(i,k)/totr(i)-C(j,k)/totr(j))**2
52          continue
54          d(i,j) = dsqrt(A*tott)
      goto 68
C Hellinger distance (as in Rao, 1995)
56    do 60 i = 1,p-1
         do 60 j = i+1,p
            A = 0.0
            do 58 k = 1,nspec
58          A = A + (dsqrt(C(i,k)/totr(i)) - dsqrt(C(j,k)/totr(j)))**2
            d(i,j) = dsqrt(A)
60    continue
      goto 68
C
C Calculation for Euclidean Distance
62    ddmin=1000000.0
      do 66 i = 1,p-1
         do 66 j = i+1,p
         A = 0.0
            do 64 k = 1,nspec
            A = A + (C(i,k) - C(j,k))**2
64          continue
      d(i,j) = dsqrt(A)
      if(d(i,j).lt.ddmin) ddmin=d(i,j)
66    continue
      goto 68
C
C d(i,j) contains at this point an upper triangular matrix of
C distances without the diagonal.  The following fills in the
C matrix and puts zeros in the diagonal.
C
68    do 70 i = 2,p
      do 70 j = 1,(i-1)
70    d(i,j) = d(j,i)
      do 72 i = 1,p
72    d(i,i) = 0.0
      return
      end

C Following are the two subroutines (from Numerical Recipes) for finding the 
C eigenvalues of a real symmetric matrix, using the Householder reduction 
C routine TRED2 followed by the routine TQLI.

      SUBROUTINE TRED2(A,N,NP,D,E)
C
C Householder reduction of a real, symmetric, N by N matrix A, stored in
C an NP by NP physical array.  On output, A is replaced by the orthogonal
C matrix Q effecting the transformation.  D returns the diagonal elements
C of the tridiagonal matrix, and E the off-diagonal elements, with E(1) = 0.
C Several statements, as noted in comments, can be omitted if only
C eigenvalues are to be found, in which case A contains no useful information
C on output.  Otherwise they are to be included.
C
      DIMENSION A(NP,NP), D(NP), E(NP)
C REAL*8 statement added for double precision (probably not needed)
      REAL*8 A,D,E,H,SCALE,F,G,HH
      
      IF (N.GT.1) THEN
         DO 18 I = N,2,-1
              L = I - 1
              H = 0.
              SCALE = 0.
              IF (L.GT.1) THEN
                 DO 11 K = 1,L
                    SCALE = SCALE + DABS(A(I,K))
11                  CONTINUE
              IF (SCALE.EQ.0) THEN
                 E(I) = A(I,L)
              ELSE
                 DO 12 K = 1,L
                    A(I,K) = A(I,K)/SCALE
                    H = H + A(I,K)**2
12               CONTINUE
                 F = A(I,L)
                 G = -SIGN(DSQRT(H),F)
                 E(I) = SCALE*G
                 H = H-F*G
                 A(I,L) = F-G
                 F = 0.
                 DO 15 J = 1,L
C Omit following line if finding only eigenvalues.
                    A(J,I) = A(I,J)/H
                    G = 0.
                    DO 13 K = 1,J
                       G = G + A(J,K) * A(I,K)
13                  CONTINUE
                    IF (L.GT.J) THEN
                       DO 14 K = J+1, L
                          G = G + A(K,J)*A(I,K)
14                     CONTINUE
                    END IF
                    E(J) = G/H
                    F = F + E(J)*A(I,J)
15                  CONTINUE
                 HH = F/(H + H)
                 DO 17 J = 1,L
                    F = A(I,J)
                    G = E(J) - HH*F
                    E(J) = G
                    DO 16 K = 1,J
                       A(J,K) = A(J,K) - F*E(K) - G*A(I,K)
16                  CONTINUE
17               CONTINUE
              END IF
              ELSE
                 E(I) = A(I,L)
              END IF
              D(I) = H
18       CONTINUE
      END IF
C Omit following line if finding only eigenvalues.
      D(1) = 0.
      E(1) = 0.
      DO 23 I = 1,N
C Delete lines from here...
         L = I-1
         IF (D(I).NE.0.) THEN
            DO 21 J = 1,L
               G = 0.
               DO 19 K = 1,L
                  G = G + A(I,K) * A(K,J)
19             CONTINUE
               DO 20 K = 1,L
                  A(K,J) = A(K,J) - G*A(K,I)
20             CONTINUE
21          CONTINUE
         END IF
C ... to here when finding only eigenvalues.
         D(I) = A(I,I)
C Also delete lines from here ....
         A(I,I) = 1.
         IF(L.GE.1) THEN
            DO 22 J = 1,L
               A(I,J) = 0.
               A(J,I) = 0.
22          CONTINUE
         END IF
C ... to here when finding only eigenvalues.
23    CONTINUE
      RETURN
      END
      
      
      SUBROUTINE TQLI (D,E,N,NP,Z)
C
C QL algorithm with implicit shifts, to determine the eigenvalues and
C eigenvectors of a real, symmetric, tridiagonal matrix previously
C reduced by TRED2 (Numerical Recipes).  D is a vector of length NP.
C On input, its first N elements are the diagonal elements of the tridiagonal
C matrix.  On output, it returns the eigenvalues.  The vector E inputs the
C subdiagonal elements of the tridiagonal matrix, with E(1) arbitrary.  On
C output, E is destroyed.  When finding only the eigenvalues, several lines
C may be omitted, as noted in the coments.  If the eigenvectors of the
C tridiagonal matrix are desired, the matrix Z (N by N matrix stored in
C NP by NP array) is input as the identity matrix.  If the eigenvectors of
C a matrix that has been reduced by TRED2 are required, then Z is input as
C the matrix output by TRED2.  In either case, the Kth column of Z returns
C the normalized eigenvector corresponding to D(K).
C
      DIMENSION D(NP), E(NP), Z(NP,NP)
C REAL*8 statement added for double precision (probably not needed)
      REAL*8 D,E,Z,DD,G,R,S,C,P,F,B

      IF (N.GT.1) THEN
         DO 11 I = 2,N
            E(I-1) = E(I)
11       CONTINUE
         E(N) = 0.
         DO 15 L = 1,N
            ITER = 0
1           DO 12 M = L, N-1
               DD = DABS(D(M)) + DABS(D(M+1))
               IF (DABS(E(M)) + DD.EQ.DD) GO TO 2
12          CONTINUE
            M = N
2           IF (M.NE.L) THEN
               IF (ITER.EQ.50) Stop 'Too many iterations in TQLI'
               ITER = ITER + 1
               G = (D(L+1)-D(L))/(2.*E(L))
               R = SQRT(G**2+1.)
               G = D(M) - D(L) + E(L)/(G + SIGN(R,G))
               S = 1.
               C = 1.
               P = 0.
               DO 14 I = M-1,L,-1
                  F = S*E(I)
                  B = C*E(I)
                  IF (DABS(F).GE.DABS(G)) THEN
                     C = G/F
                     R = DSQRT(C**2+1.)
                     E(I+1) = F*R
                     S = 1./R
                     C = C*S
                  ELSE
                     S = F/G
                     R = DSQRT(S**2+1.)
                     E(I+1) = G*R
                     C = 1./R
                     S = S*C
                  END IF
                  G = D(I+1) - P
                  R = (D(I) - G)*S+2.*C*B
                  P = S*R
                  D(I+1) = G + P
                  G = C*R - B
C Omit lines from here ...
                  DO 13 K = 1,N
                     F = Z(K,I+1)
                     Z(K,I+1) = S*Z(K,I) + C*F
                     Z(K,I) = C*Z(K,I) - S*F
13                CONTINUE
C ... to here when finding only eigenvalues.
14             CONTINUE
               D(L) = D(L) - P
               E(L) = G
               E(M) = 0.
               GO TO 1
            END IF
15          CONTINUE
      END IF
      RETURN
      END


C Following are three subroutines (from Numerical Recipes, Section 11.6) for 
C finding eigenvalues of a real non-symmetric matrix: BALANC, ELMHES and HQR.

      Subroutine Balanc(a,n,np)
C
C Given an N by N matrix A stored in an array of physical dimensions NP by NP, 
C this routine replaces it by a balanced matrix with identical eigenvalues. A 
C symmetric matrix is already balanced and is unaffected by this procedure. 
C The parameter RADIX should be the machine's floating point radix.
C Balancing is a procedure with of order N^2 operations. The time taken should
C be never more than a few percent of the total time required to find the 
C eigenvalues. From Numerical recipes, Section 11.5.
C
      Real*8 a(np,np),c,f,g,r,s,radix,sqrdx
      radix=2.0
      sqrdx=radix**2
    1 continue
      last = 1
      do 14 i=1,n
         c = 0.0
         r = 0.0
         do 11 j = 1,n
            if (j.ne.i) then
               c = c+dabs(a(j,i))
               r = r+dabs(a(i,j))
               endif
   11    continue
         if ((c.ne.0.0).and.(r.ne.0.0)) then
            g = r/radix
            f = 1.0
            s = c+r
    2       if(c.lt.g) then
               f = f*radix
               c = c*sqrdx
               goto 2
               endif
            g = r*radix
    3       if(c.gt.g) then
               f = f/radix
               c = c/sqrdx
               goto 3
               endif
            if (((c+r)/f).lt.(0.95*s)) then
               last = 0
               g = 1.0/f
               do 12 j=1,n
   12          a(i,j) = a(i,j)*g
               do 13 j=1,n
   13          a(j,i) = a(j,i)*f
               endif
            endif
   14 continue
      if(last.eq.0) goto 1
      return
      end

      Subroutine Elmhes(a,n,np)
C
C Reduction to Hessenberg form by the elimination method. The real nonsymmetric
C N by N matrix A, stored in an array of physical dimension NP by NP, is  
C replaced by an upper Hessenberg matrix with identical eigenvalues. 
C Recommended, but not required, if that this routine be preceded by BALANC.
C On output, the Hessenberg matrix is in elements A(i,j) with i <= j+1. Elements
C with i > j+1 are to be thought of as zero, but are returned with random 
C values. From Numerical recipes, Section 11.5.
C
      Real*8 a(np,np),x,y
      if(n.gt.2) then
      do 17 m=2,n-1
         x = 0.0
         i = m
C Find the pivot
         do 11 j = m,n
            if (dabs(a(j,m-1)).gt.dabs(x)) then
               x = a(j,m-1)
               i = j
               endif
   11    continue
C Interchange rows and columns
         if (i.ne.m) then
            do 12 j = m-1,n
               y = a(i,j)
               a(i,j) = a(m,j)
               a(m,j) = y
   12       continue
            do 13 j = 1,n
               y = a(j,i)
               a(j,i) = a(j,m)
               a(j,m) = y
   13       continue
         endif
C Carry out the elimination
         if (x.ne.0.0) then
            do 16 i = m+1,n
               y = a(i,m-1)
               if (y.ne.0.0) then
                  y = y/x
                  a(i,m-1) = y
                  do 14 j = m,n
   14             a(i,j) = a(i,j)-y*a(m,j)
                  do 15 j=1,n
   15             a(j,m) = a(j,m)+y*a(j,i)
                  endif
   16       continue
            endif
   17 continue
      endif
      return
      end

      Subroutine Hqr(a,n,np,wr,wi)
C
C Finds all eigenvalues of an N by N upper Hessenberg matrix A that is stored in
C an NP by NP array. On input A can be exactly as output from ELMHES. On output,
C it is destroyed. The real and imaginary parts of the eigenvalues are returned
C in WR and WI respectively.
C
      Real*8 a(np,np),wr(np),wi(np),anorm,p,q,r,s,t,u,v,w,x,y,z
C Compute matrix norm for possible use in locating single small subdiagonal 
C elements
      anorm = dabs(a(1,1))
      do 12 i = 2,n
      do 11 j = i-1,n
   11 anorm = anorm+dabs(a(i,j))
   12 continue
      nn = n
      t = 0.0
C Begin search for next eigenvalue
    1 if (nn.ge.1) then
         its = 0
C Begin iteration: look for single small subdiagonal element
    2    do 13 L = nn,2,-1
            s = dabs(a(L-1,L-1))+dabs(a(L,L))
            if (s.eq.0.0) s = anorm
            if ((dabs(a(L,L-1))+s).eq.s) goto 3
   13    continue
         L = 1
    3    x = a(nn,nn)
         if (L.eq.nn) then
            wr(nn) = x+t
            wi(nn) = 0.0
            nn = nn-1
         else
            y = a(nn-1,nn-1)
            w = a(nn,nn-1)*a(nn-1,nn)
            if (L.eq.nn-1) then
               p = 0.5*(y-x)
               q = p**2+w
               z = dsqrt(dabs(q))
               x = x+t
               if (q.ge.0.0) then
                  z = p+dsign(z,p)
                  wr(nn) = x+z
                  wr(nn-1) = wr(nn)
                  if (z.ne.0.0) wr(nn) = x-w/z
                  wi(nn) = 0.0
                  wi(nn-1) = 0.0
               else
                  wr(nn) = x+p
                  wr(nn-1) = wr(nn)
                  wi(nn) = z
                  wi(nn-1) = -z
               endif
               nn = nn-2
            else
               if (its.eq.30) Stop 'Too many iterations in HQR'
               if ((its.eq.10).or.(its.eq.20)) then
                  t = t+x
                  do 14 i = 1,nn
   14             a(i,i) = a(i,i)-x
                  s = dabs(a(nn,nn-1))+dabs(a(nn-1,nn-2))
                  x = 0.75*s
                  y = x
                  w = -0.4375*s**2
               endif
               its = its+1
               do 15 m = nn-2,L,-1
                  z = a(m,m)
                  r = x-z
                  s = y-z
                  p = (r*s-w)/a(m+1,m)+a(m,m+1)
                  q = a(m+1,m+1)-z-r-s
                  r = a(m+2,m+1)
                  s = dabs(p)+dabs(q)+dabs(r)
                  p = p/s
                  q = q/s
                  r = r/s
                  if (m.eq.L) goto 4
                  u = dabs(a(m,m-1))*(dabs(q)+dabs(r))
C Mod. PhC 17/05/00
C Computation of v split in two lines to address 72-char limit
                  v = dabs(p)
                  v = v*(dabs(a(m-1,m-1))+dabs(z)+dabs(a(m+1,m+1)))
                  if ((u+v).eq.v) goto 4
   15          continue
    4          do 16 i = m+2,nn
                  a(i,i-2) = 0.0
                  if  (i.ne.(m+2)) a(i,i-3) = 0.0
   16          continue
               do 19 k = m,nn-1
                  if (k.ne.m) then
                     p = a(k,k-1)
                     q = a(k+1,k-1)
                     r = 0.0
                     if (k.ne.(nn-1)) r = a(k+2,k-1)
                     x = dabs(p)+dabs(q)+dabs(r)
                     if (x.ne.0.0) then
                        p = p/x
                        q = q/x
                        r = r/x
                     endif
                  endif
                  s = dsign(dsqrt(p**2+q**2+r**2),p)
                  if (s.ne.0.0) then
                     if (k.eq.m) then
                        if (L.ne.m) a(k,k-1) = -a(k,k-1)
                     else
                        a(k,k-1) = -s*x
                     endif
                     p = p+s
                     x = p/s
                     y = q/s
                     z = r/s
                     q = q/p
                     r = r/p
C Row modification
                     do 17 j = k,nn
                        p = a(k,j)+q*a(k+1,j)
                        if (k.ne.(nn-1)) then
                           p = p+r*a(k+2,j)
                           a(k+2,j) = a(k+2,j)-p*z
                        endif
                        a(k+1,j) = a(k+1,j)-p*y
                        a(k,j) = a(k,j)-p*x
   17                continue
C Column modification
                     do 18 i = L,min(nn,k+3)
                        p = x*a(i,k)+y*a(i,k+1)
                        if (k.ne.(nn-1)) then
                           p = p+z*a(i,k+2)
                           a(i,k+2) = a(i,k+2)-p*r
                        endif
                        a(i,k+1) = a(i,k+1)-p*q
                        a(i,k) = a(i,k)-p
   18                continue
               endif
   19          continue
               goto 2
            endif
         endif
         goto 1
      endif
      return
      end
